# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import constants
from oslo_log import log
from tempest.common.utils.linux import remote_client
from tempest.common import waiters
from tempest import exceptions
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from whitebox_neutron_tempest_plugin.tests.scenario import \
    test_sriov_provider_network


CONF = config.CONF
WB_CONF = config.CONF.whitebox_neutron_plugin_options
LOG = log.getLogger(__name__)


class RoutedProviderNetworkSriovBaseTest(
        test_sriov_provider_network.ProviderNetworkSriovBaseTest):
    """Base class for tests using provisioning networks
    These tests need the same deployment:
    - datacentre with controllers and 1 compute over subnet 1
    - leaf1 1 computes over subnet 2
    and the next preconfiguration:
    - Admin user is needed to create ports on the existing provisioning network
    - 2 network segments called segment1, segment2
    - 2 subnets segment1(subnet1), segment2(subnet2)
    - 2 AZs segment1(central site), segment2(leaf1)
    """

    @classmethod
    def skip_checks(cls):
        super(RoutedProviderNetworkSriovBaseTest, cls).skip_checks()
        if not WB_CONF.provroutednetwork_support:
            raise cls.skipException(
                "Skipped since according to provroutednetwork_support config "
                "value Provider Routed Networks are not supported")

    def _test_create_instance_with_network_port(self, port_type,
                                                subnet_id='1',
                                                vm_name=None,
                                                reuse_port=False):
        # Create vm in segmentX, az segmentX, subnet segmentX
        # Vm created directly over the external network
        # az segmentX
        # X=subnet_id
        security_groups, port, user_data, config_drive = \
            self._create_network_port(port_type,
                                      reuse_port=reuse_port,
                                      use_provider_net=True,
                                      subnet_id=subnet_id)
        vm_name = vm_name or "vm-" + port_type + "-" + self._testMethodName
        subnet_name = 'segment' + subnet_id
        self._create_server(
            security_groups=security_groups, networks=[port],
            user_data=user_data, config_drive=config_drive,
            availability_zone=subnet_name, name=vm_name)

    def _test_instance_with_one_port_prov_net(self, port_type, subnet_id='1'):
        # Create vm in segmentX, az segmentX, subnet segmentX
        # Vm created over a pre created port in subnet
        # segmentX and az segmentX
        # X=subnet_id
        self._test_create_instance_with_network_port(port_type, subnet_id)
        self.check_connectivity(self.ports[-1]['fixed_ips'][0]['ip_address'],
                                self.username,
                                self.keypair['private_key'], ssh_timeout=60)
        self.check_port_status(port_type)


class OneServerRoutedProvNetSriovTest(RoutedProviderNetworkSriovBaseTest):
    """Class for SRIOV tests creating one VM with one SRIOV port on routed
    provider networks
    These tests can be executed without needing to cleanup VMs or ports between
    them
    """

    @decorators.idempotent_id('57bc00f2-d166-41c1-8a87-600bb087f248')
    def test_create_instance_with_normal_port_prov_net_seg1(self):
        # Create vm in datacentre network, segment1, az segment1,
        # subnet segment1
        # Vm created over a pre created port in subnet
        # segment1 and az segment1
        self._test_instance_with_one_port_prov_net('normal', '1')

    @decorators.idempotent_id('d6356ffd-e521-48d2-bc69-13f2cd72690a')
    def test_create_instance_with_normal_port_prov_net_seg2(self):
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        self._test_instance_with_one_port_prov_net('normal', '2')

    @decorators.idempotent_id('aea44620-6e01-4cf1-8b9c-9171717092b0')
    def test_create_instance_with_direct_port_prov_net_seg1(self):
        # Create vm in datacentre network, segment1, az segment1,
        # subnet segment1
        # Vm created over a pre created port in subnet
        # segment1 and az segment1
        self._test_instance_with_one_port_prov_net('direct', '1')

    @decorators.idempotent_id('17e96d5d-c5b6-4731-8027-7810a0761da3')
    def test_create_instance_with_direct_port_prov_net_seg2(self):
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        self._test_instance_with_one_port_prov_net('direct', '2')

    @decorators.idempotent_id('6d2f7a06-6f41-42e7-a239-f722914914da')
    def test_create_instance_with_macvtap_port_prov_net_seg1(self):
        # Create vm in datacentre network, segment1, az segment1,
        # subnet segment1
        # Vm created over a pre created port in subnet
        # segment1 and az segment1
        self._test_instance_with_one_port_prov_net('macvtap', '1')

    @decorators.idempotent_id('955c6502-9cc6-4c7e-bb12-726be03a2c75')
    def test_create_instance_with_macvtap_port_prov_net_seg2(self):
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        self._test_instance_with_one_port_prov_net('macvtap', '2')


class OneServerPfRoutedProvNetSriovTest(RoutedProviderNetworkSriovBaseTest):
    """The following test need to be executed isolated (VM and port cleanup is
    needed before and after its execution)
    """

    @decorators.idempotent_id('18e1fc90-440a-43ea-b269-524064332bb0')
    def test_create_instance_with_direct_physical_port_prov_net_seg1(self):
        # Create vm in datacentre network, segment1, az segment1,
        # subnet segment1
        # Vm created over a pre created port in subnet
        # segment1 and az segment1
        self._test_instance_with_one_port_prov_net('direct-physical', '1')

    @decorators.idempotent_id('a9fa54d0-ec91-4d46-978d-48e3cef4492e')
    def test_create_instance_with_direct_physical_port_prov_net_seg2(self):
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        self._test_instance_with_one_port_prov_net('direct-physical', '2')


class RebootServerRoutedProvNetSriovRoutedTest(
        RoutedProviderNetworkSriovBaseTest):
    """Class for SRIOV test that creates VM instances with different SRIOV port
    types, reboot those instances and checks their behavior is correct after
    reboot
    """

    @decorators.idempotent_id('43482b46-bb26-422f-96cd-99508eb30a75')
    def test_reboot_vm_with_sriov_port_prov_net_seg2(self):
        # Create vm in datacentre network, segment1, az segment1,
        # subnet segment1
        # Vm created over a pre created port in subnet
        # segment1 and az segment1
        # Reboot the vm
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        # Reboot the vm
        cont = 0
        for port_type in ('direct', 'direct-physical'):
            cont += 1
            self._test_instance_with_one_port_prov_net(port_type, str(cont))
            for reboot_type in ('SOFT', 'HARD'):
                self.servers_client.reboot_server(self.servers[-1]['id'],
                                                  type=reboot_type)
                waiters.wait_for_server_status(self.servers_client,
                                               self.servers[-1]['id'],
                                               constants.SERVER_STATUS_ACTIVE)
                self.check_connectivity(
                    self.ports[-1]['fixed_ips'][0]['ip_address'],
                    self.username,
                    self.keypair['private_key'])
                self.check_port_status(port_type)


class ReusePortRoutedProvNetSriovRoutedTest(
        RoutedProviderNetworkSriovBaseTest):
    """Class for SRIOV test that creates VM instances with different SRIOV port
    types, delete those instances and reuse the existing SRIOV ports on new VM
    instances
    """

    @decorators.idempotent_id('eab71091-8a5f-4609-b259-91de74ad8744')
    def test_reuse_sriov_port_prov_net_seg2(self):
        # TODO(eolivare): only VF ports supported at this moment
        # for port_type in ('direct', 'direct-physical'):
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        # Delete vm
        # Create vm reusing the port
        for port_type in ('direct',):
            self._test_instance_with_one_port_prov_net(port_type, '2')
            old_port_id = self.ports[-1]['id']
            old_num_ports = len(self.ports)
            # delete the server and create a new server attached to the
            # existing port
            self.servers_client.delete_server(self.servers[-1]['id'])
            waiters.wait_for_server_termination(self.servers_client,
                                                self.servers[-1]['id'])
            del self.servers[-1]
            # validate port is not reachable anymore
            self.assertRaises(
                lib_exc.SSHTimeout,
                self.check_connectivity,
                host=self.ports[-1]['fixed_ips'][0]['ip_address'],
                ssh_user=self.username,
                ssh_key=self.keypair['private_key'],
                ssh_timeout=10)

            # create a new server reusing the existing port
            self._test_create_instance_with_network_port(
                port_type, '2', reuse_port=True)
            waiters.wait_for_server_status(self.servers_client,
                                           self.servers[-1]['id'],
                                           constants.SERVER_STATUS_ACTIVE)
            # check no new ports have been created
            self.assertEqual(old_num_ports, len(self.ports))
            self.assertEqual(old_port_id, self.ports[-1]['id'])
            # validate port is reachable again
            self.check_connectivity(
                self.ports[-1]['fixed_ips'][0]['ip_address'],
                self.username,
                self.keypair['private_key'])
            self.check_port_status(port_type)

    @decorators.idempotent_id('b7bed0e4-9c34-4bf1-b411-246345ef7338')
    def test_detach_attach_sriov_port_prov_net_seg2(self):
        interfaces_client = self.os_adm.interfaces_client
        # TODO(eolivare): only VF ports supported at this moment
        # for port_type in ('direct', 'direct-physical'):
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        # Remove interface from vm
        # Create vm reusing the port
        for port_type in ('direct',):
            self._test_instance_with_one_port_prov_net(port_type, '2')
            old_port_id = self.ports[-1]['id']
            old_num_ports = len(self.ports)
            # remove port from server
            interfaces_client.delete_interface(self.servers[-1]['id'],
                                               self.ports[-1]['id'])
            # validate port is not reachable anymore
            self.assertRaises(
                lib_exc.SSHTimeout,
                self.check_connectivity,
                host=self.ports[-1]['fixed_ips'][0]['ip_address'],
                ssh_user=self.username,
                ssh_key=self.keypair['private_key'],
                ssh_timeout=10)

            # reattach port on a new server (reattaching to a running VM is
            # not supported)
            self._test_create_instance_with_network_port(
                port_type, '2', reuse_port=True)
            waiters.wait_for_server_status(self.servers_client,
                                           self.servers[-1]['id'],
                                           constants.SERVER_STATUS_ACTIVE)
            # check no new ports have been created
            self.assertEqual(old_num_ports, len(self.ports))
            self.assertEqual(old_port_id, self.ports[-1]['id'])
            # validate port is reachable again
            self.check_connectivity(
                self.ports[-1]['fixed_ips'][0]['ip_address'],
                self.username,
                self.keypair['private_key'])
            self.check_port_status(port_type)


class NegativeProvRoutedNetSriovRoutedTest(RoutedProviderNetworkSriovBaseTest):
    """Assuming we have hosts with one SRIOV port and it is already occupied,
    then creation of VM with VF or PF port should fail

    We create a VM with PF port on each compute node in order to occupy SRIOV
    port; creating VM with PF or VF should fail as there is no available port
    """

    @decorators.idempotent_id('94dd5592-3cc9-4636-bad7-077c0b6b5558')
    def test_try_create_instance_with_sriov_port_seg2(self):
        # Create vm in leaf1 network, segment2, az segment2,
        # subnet segment2
        # Vm created over a pre created port in subnet
        # segment2 and az segment2
        # Try to create a second vm with vp and pf
        self._test_create_instance_with_network_port('direct-physical', '2')

        for port_type in ('direct-physical', 'direct'):
            vm_name = 'vm-%s-error' % port_type
            self.assertRaises(
                exceptions.BuildErrorException,
                self._test_create_instance_with_network_port,
                port_type=port_type, subnet_id='2',
                vm_name=vm_name)


class TwoServersProvNetSriovRoutedTest(RoutedProviderNetworkSriovBaseTest):
    """Base class for SRIOV tests creating two VMs with one SRIOV port on
    routed provider networks
        # Create 2 vm:
        # 1 segmentX1, az segmentX1, subnet segmentX1
        # 2 segmentX2, az segmentX2, subnet segmentX2
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        # X1=subnet_id1
        # X2=subnet_id2
    """

    def _test_create_two_instances_with_ports_seg(self,
                                                  port_type1, segment1,
                                                  port_type2, segment2):
        self._test_create_instance_with_network_port(port_type1, segment1)
        ip1 = self.ports[0]['fixed_ips'][0]['ip_address']
        self._test_create_instance_with_network_port(port_type2, segment2)
        ip2 = self.ports[1]['fixed_ips'][0]['ip_address']
        for ip in (ip1, ip2):
            self.check_connectivity(ip,
                                    self.username,
                                    self.keypair['private_key'])
        server_ssh_client = remote_client.RemoteClient(
            ip1,
            self.username,
            pkey=self.keypair['private_key'], server=self.servers[0])
        server_ssh_client.ping_host(ip2)


class TwoPortsTwoServersRoutedProvNetSriovTest(
        TwoServersProvNetSriovRoutedTest):
    """The following tests can be executed without needing to cleanup between
    them because only normal and Vf ports are created
    """

    @decorators.idempotent_id('4523524a-b70e-4f44-9984-f6e622e652d1')
    def test_create_instances_with_normal_normal_ports_seg1_seg2(self):
        # Create 2 vm:
        # 1 segment1, az segment1, subnet segment1
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the different segment(central site and leaf)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('normal', '1',
                                                       'normal', '2')

    @decorators.idempotent_id('b5a96f9f-68b7-4559-b98f-b2c1b3f3a940')
    def test_create_instances_with_normal_normal_ports_seg2(self):
        # Create 2 vm:
        # 1 segment2, az segment2, subnet segment2
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the same segment and compute(leaf1)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('normal', '2',
                                                       'normal', '2')

    @decorators.idempotent_id('6cbfaf93-b3d0-4a52-83c2-69a387e815e1')
    def test_create_instances_with_normal_vf_ports_seg1_seg2(self):
        # Create 2 vm:
        # 1 segment1, az segment1, subnet segment1
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the different segment(central site and leaf)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('direct', '1',
                                                       'normal', '2')

    @decorators.idempotent_id('b45d44e4-aa13-484a-86d8-19482bdd9c25')
    def test_create_instances_with_normal_vf_ports_seg2(self):
        # Create 2 vm:
        # 1 segment2, az segment2, subnet segment2
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the same segment and compute(leaf1)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('direct', '2',
                                                       'normal', '2')

    @decorators.idempotent_id('bec613f7-57ab-44ea-879f-e6a21647493a')
    def test_create_instances_with_vf_vf_ports_seg1_seg2(self):
        # Create 2 vm:
        # 1 segment1, az segment1, subnet segment1
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the different segment(central site and leaf)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('direct', '1',
                                                       'direct', '2')


class PfAndPfServersProvNetSriovTest(
        TwoServersProvNetSriovRoutedTest):
    """The following test need to be executed isolated (VM and port cleanup is
    needed before and after its execution)
    """

    @decorators.idempotent_id('128f6070-b0fc-4dc3-a815-2aeaf43f2815')
    def test_create_instances_with_pf_pf_ports_seg1_seg2(self):
        # Create 2 vm:
        # 1 segment1, az segment1, subnet segment1
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the different segment(central site and leaf)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('direct-physical', '1',
                                                       'direct-physical', '2')


class VfAndPfServersRoutedProvNetSriovTest(
        TwoServersProvNetSriovRoutedTest):
    """The following test need to be executed isolated (VM and port cleanup is
    needed before and after its execution)
    """

    @decorators.idempotent_id('7eecb5f4-2dde-4cf6-bf60-e3663bade307')
    def test_create_instances_with_vf_pf_ports_seg1_seg2(self):
        # Create 2 vm:
        # 1 segment1, az segment1, subnet segment1
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the different segment(central site and leaf)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('direct-physical', '1',
                                                       'direct', '2')


class NormalAndPf1SegServersRoutedProvNetSriovTest(
        TwoServersProvNetSriovRoutedTest):
    """The following test need to be executed isolated (VM and port cleanup is
    needed before and after its execution)
    """

    @decorators.idempotent_id('9de6e9c3-b1e1-4509-b97b-dd9800fee2b9')
    def test_create_instances_with_normal_pf_ports_seg2(self):
        # Create 2 vm:
        # 1 segment2, az segment2, subnet segment2
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the same segment and compute(leaf1)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('direct-physical', '2',
                                                       'normal', '2')


class NormalAndPfServersRoutedProvNetSriovTest(
        TwoServersProvNetSriovRoutedTest):
    """The following test need to be executed isolated (VM and port cleanup is
    needed before and after its execution)
    """

    @decorators.idempotent_id('5360a20f-9168-494a-9c32-72841a9f7563')
    def test_create_instances_with_normal_pf_ports_seg1_seg2(self):
        # Create 2 vm:
        # 1 segment1, az segment1, subnet segment1
        # 2 segment2, az segment2, subnet segment2
        # 2 vms in the different segment(central site and leaf)
        # Vms created over a pre created port
        # az segmentX, subnet segmentX
        self._test_create_two_instances_with_ports_seg('direct-physical', '1',
                                                       'normal', '2')
